<?php

declare(strict_types = 1);

namespace Drupal\authorization_group\Plugin\authorization\consumer;

use Drupal\authorization\Consumer\ConsumerPluginBase;
use Drupal\Component\Utility\Html;
use Drupal\Core\Form\FormStateInterface;
use Drupal\group\Entity\Group;
use Drupal\group\Entity\GroupRole;
use Drupal\group\GroupMembershipLoaderInterface;
use Drupal\user\UserInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Using Authorization to assign permissions to Groups.
 *
 * Groups are provided by the Group module.
 *
 * @AuthorizationConsumer(
 *   id = "authorization_group",
 *   label = @Translation("Groups")
 * )
 */
class GroupConsumer extends ConsumerPluginBase {

  /**
   * Allow consumer target creation.
   *
   * @var bool
   */
  protected $allowConsumerTargetCreation = FALSE;

  /**
   * The membership loader service.
   *
   * @var \Drupal\group\GroupMembershipLoaderInterface
   */
  protected $membershipLoader;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, array $plugin_definition, GroupMembershipLoaderInterface $membership_loader) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->membershipLoader = $membership_loader;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('group.membership_loader')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['delete_membership'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Remove the user from the Group, if they no longer have any roles'),
      '#default_value' => $this->configuration['delete_membership'],
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function buildRowForm(array $form, FormStateInterface $form_state, $index): array {
    $row = [];
    $mappings = $this->configuration['profile']->getConsumerMappings();

    // Load all available groups, traverse through each of them, load all
    // available roles attached to that group, and assemble an array with the
    // information.
    // @todo Should we have a filter on this?
    $groups = \Drupal::entityQuery('group')->execute();
    $group_options = ['none' => $this->t('- None -')];

    /** @var \Drupal\group\Entity\GroupInterface $group */
    foreach (Group::loadMultiple($groups) as $group_id => $group) {
      $group_name = Html::escape($group->label());
      $group_type = $group->getGroupType();

      foreach ($group_type->getRoles() as $role) {
        if (/*($role->isMember() || $role->isAnonymous()) && */ $role->isInternal()) {
          $group_options[$group_name][$group_id . '--' . $role->id()] = $role->label();
        }
      }
    }
    $row['group'] = [
      '#type' => 'select',
      '#title' => $this->t('Group and Role'),
      '#options' => $group_options,
      '#default_value' => $mappings[$index]['group'],
      '#description' => 'Choose the Group to apply to the user.',
    ];
    return $row;
  }

  /**
   * {@inheritdoc}
   */
  public function createConsumerTarget(string $consumer): void {
    // @todo Implement createConsumerTarget() method.
    // We are not currently allowing auto-recreation of AD groups in Drupal.
    // No work to do.
  }

  /**
   * {@inheritdoc}
   */
  public function getTokens(): array {
    $tokens = parent::getTokens();

    // Reset the tokens for plurality.
    $tokens['@' . $this->getType() . '_namePlural'] = $this->label();
    $tokens['@' . $this->getType() . '_name'] = 'Group';
    return $tokens;
  }

  /**
   * {@inheritdoc}
   */
  public function grantSingleAuthorization(UserInterface $user, $consumerMapping): void {
    if (!empty($consumerMapping) && $consumerMapping !== 'none' && $user->id()) {

      [$group_id, $role_id] = explode('--', $consumerMapping);
      $group = Group::load($group_id);
      $role = GroupRole::load($role_id);

      $group->addMember($user);

      $assign_new_role = TRUE;

      $membership_roles = [];

      // Go through all the roles that the user currently have, within the
      // group.
      // If the user isn't assigned to the role yet, the assign it. Otherwise,
      // leave the roles as they are.
      $memberships = $group->getContentByEntityId('group_membership', $user->id());
      if (!empty($memberships) && \is_array($memberships) && $role) {
        /** @var \Drupal\group\Entity\GroupContent $membership */
        foreach ($memberships as $group_content_id => $membership) {
          $group_roles = $membership->get('group_roles');
          /** @var \Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItem $group_role */
          foreach ($group_roles as $index => $group_role) {
            $membership_roles[] = ['target_id' => $group_role->target_id];
            if ($group_role->target_id == $role_id) {
              $assign_new_role = FALSE;
            }
          }

          if ($assign_new_role) {
            $membership_roles[] = ['target_id' => $role_id];
            $this->messenger()->addStatus($this->t('You have been granted the role %role in the %group group', [
              '%role' => $role->label(),
              '%group' => $group->label(),
            ]));
            $membership->set('group_roles', $membership_roles);
            // @todo Handle potential exception here.
            $membership->save();
          }
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function revokeGrants(UserInterface $user, array $context): void {
    if (!$user->id()) {
      return;
    }

    // Find the granted groups and roles from the applied grants (context).
    $granted_group_roles = [];
    foreach ($context as $mapping) {
      [$group_id, $role_id] = explode('--', $mapping);
      $granted_group_roles[$group_id][] = $role_id;
    }

    // Load all of the user's memberships.
    $memberships = $this->membershipLoader->loadByUser($user);

    foreach ($memberships as $membership) {
      $group = $membership->getGroup();

      // Wrapped membership content.
      $membershipContent = $membership->getGroupContent();

      if (!isset($granted_group_roles[$group->id()])) {
        // Entire group was removed, so either remove all roles or delete membership.
        if ($this->configuration['delete_membership']) {
          $membershipContent->delete();
        }
        else {
          $membershipContent->set('group_roles', []);
          $membershipContent->save();
        }
      }
      else {
        // The group exists in the grants, so check all of the roles.
        $roles = $membership->getRoles();
        $roles_to_keep = [];

        foreach ($roles as $role) {
          if (in_array($role->id(), $granted_group_roles[$group->id()])) {
            // Role exists in grants, so keep it.
            $roles_to_keep[] = ['target_id' => $role->id()];
          }
          else {
            $this->messenger()->addStatus($this->t('Your %role role, in the %group group, has been revoked', [
              '%role' => $role->label(),
              '%group' => $group->label(),
            ]));
          }
        }

        // Update the membership's roles.
        $membershipContent->set('group_roles', $roles_to_keep);
        $membershipContent->save();
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    parent::submitConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitRowForm(array &$form, FormStateInterface $form_state): void {
    $values = $form_state->getValues();

    $consumer_mappings = [];
    foreach ($values as $key => $value) {
      $consumer_mappings[] = $value['consumer_mappings'];
    }

    $form_state->setValue('consumer_mappings', $consumer_mappings);

    parent::submitRowForm($form, $form_state);
  }

}
